@@ -1,10 +1,17 @@
 /*
- * Copyright (c) 2005 QUALCOMM Incorporated.  All rights reserved.
+ * Copyright (c) 2008 QUALCOMM Incorporated.  All rights reserved.
  * The file License.txt specifies the terms for use, modification,
  * and redistribution.
  *
  * Revisions:
  *
+ *     04/18/08  [rcg]
+ *              - Fix strlcpy off-by-one error when source exceeds dest.
+ *              - Prevent strlcat from examining destination past
+ *                buffer size.
+ *              - Guard strlcpy and strlcat so they aren't compiled if
+ *                the platform supplies them.
+ *
  *     11/03/00  [rcg]
  *              - Ensure that string_copy null-terminates.
  *
@@ -37,6 +44,7 @@
 
 
 
+#ifndef    HAVE_STRLCPY
 /*
  * Function:        strlcpy
  *
@@ -52,27 +60,27 @@
  *                  is always null-terminated (as long as its size > 0).
  */
 
-size_t
-strlcpy ( char *strdest, const char *strsource, size_t bufsize )
+size_t strlcpy ( char *strdest, const char *strsource, size_t bufsize )
 {
     register char       *dst = strdest;
     register const char *src = strsource;
-    register size_t      len = 0;
+    register size_t      len = bufsize;
 
     /*
      * Copy characters as long as they fit 
      */
-    for ( len = bufsize; len > 0; len-- )
+    while ( len > 1 ) /* save one for the terminating null */
     {
         *dst++ = *src;
         if ( *src++ == '\0' )
             break;
+        len--;
     }
     
     /*
      * Add null if we stopped early
      */
-    if ( len == 0 && bufsize != 0 )
+    if ( len == 1 )
     {
         *dst = '\0';
     }
@@ -85,8 +93,11 @@ strlcpy ( char *strdest, const char *strsource, size_t bufsize )
         src++; /* make sure we are at end of source string */ 
     return ( src - strsource );
 }
+#endif /* HAVE_STRLCPY */
 
 
+
+#ifndef    HAVE_STRLCAT
 /*
  * Function:        strlcat
  *
@@ -107,21 +118,33 @@ strlcpy ( char *strdest, const char *strsource, size_t bufsize )
  *                  null-terminated (as long as its size > 0).
  */
 
-size_t
-strlcat ( char *strdest, const char *strsource, size_t bufsize )
+size_t strlcat ( char *strdest, const char *strsource, size_t bufsize )
 {
     register char       *dst    = strdest;
     register const char *src    = strsource;
     register size_t      avail  = 0;
     size_t               dstlen = 0;
-    size_t               tmp    = 0;
+    size_t               tmp    = bufsize;
 
 
-    if ( bufsize == 0 )
-        return strlen(strdest) + strlen(strsource);
+    /*
+     * Get length of strdest without going past bufsize
+     */
+    while ( tmp-- != 0 && *dst != '\0' )
+    {
+        dst++;
+    }
+    dstlen = dst - strdest;
 
-    dstlen = strlen ( strdest );
-    dst   += dstlen; /* 'Append at the end', Alice might have said. */
+    /*
+     * If no space available, don't copy anything, just return
+     * theoretical length of string if we had copied everything.
+     */
+    tmp = bufsize - dstlen;
+    if ( tmp == 0 )
+        return ( dstlen + strlen(strsource) );
+
+    /* 'Append at the end', Alice might have said. */
 
     /* 
      * Calculate available space in buffer.
@@ -131,13 +154,6 @@ strlcat ( char *strdest, const char *strsource, size_t bufsize )
         tmp = bufsize -1;
     avail  = bufsize - tmp;
 
-    /*
-     * If no space available, don't copy anything, just return
-     * theoretical length of string if we had copied everything.
-     */
-    if ( avail == 0 )
-        return dstlen + strlen(strsource);
-
     /*
      * Copy what we can.
      */
@@ -156,6 +172,7 @@ strlcat ( char *strdest, const char *strsource, size_t bufsize )
     *dst = '\0';
     return dstlen + ( src - strsource ) + strlen ( src );
 }
+#endif /* HAVE_STRLCAT */
 
 
 /*
