@@ -1,10 +1,25 @@
 /* 
- * Copyright (c) 2003 QUALCOMM Incorporated.  All rights reserved.
+ * Copyright (c) 2008 QUALCOMM Incorporated.  All rights reserved.
  * The file license.txt specifies the terms for use, modification,
  * and redistribution.
  *
  * Revisions:
  *
+ *     04/25/05  [rcg]
+ *              - Fix '-V' for standalone.
+ *
+ *     12/15/04  [rcg]
+ *              - Log facility set on command line now applies to
+ *                daemon as well.  [Thanks to Helge Oldach]
+ *
+ *     07/23/04  [rcg]
+ *              - when configured as --with-pam and building for 
+ *                Darwin, include <pam/pam_appl.h> instead of 
+ *                <security/pam_appl.h> (otherwise build fails);
+ *                also work around PAM bug;
+ *                also open stdout and stderr write-only instead of read-only
+ *                (thanks to Kyle McKay for the patch)
+ *
  *     11/27/01  [rcg]
  *              - Worked around a bug (since fixed) in OpenBSD by
  *                setting file descriptor to be blocking *before*
@@ -107,6 +122,13 @@
 #  include <sys/fcntl.h>
 #endif /* HAVE_SYS_FCNTL_H */
 
+#ifdef USE_PAM
+#  ifdef    HAVE_PAM_PAM_APPL_H
+#    include <pam/pam_appl.h>
+#  elif defined HAVE_SECURITY_PAM_APPL_H
+#    include <security/pam_appl.h>
+#  endif /* HAVE_PAM_PAM_APPL_H */
+#endif /* USE_PAM */
 
 #ifndef  STANDALONE
 
@@ -114,11 +136,25 @@
  * When not in standalone mode, main() is a jacket for qpopper().
  */
 int
-main ( argc, argv )
-int argc;
-char *argv[];
+main ( int argc, char *argv[] )
 {
+
+#ifdef USE_PAM
+    {
+        /* Workaround a bug where if we wait until after all the daemonizing
+         * closing and opening of file handles, etc., pam_start may appear
+         * to succeed, but will fail to authenticate the user.
+         */
+        pam_handle_t    *pamdummy    = NULL;
+        struct pam_conv pamdummyconv = {NULL, NULL};
+        if ( pam_start ( USE_PAM, NULL, &pamdummyconv, &pamdummy) == PAM_SUCCESS)
+            pam_end ( pamdummy, 0 );
+    }
+#endif /* USE_PAM */
+
     if ( argc >= 2 && ( strncmp ( argv[1], "-v",  2 ) == 0 ||
+                        strncmp ( argv[1], "-V",  2 ) == 0 ||
+                        strncmp ( argv[1], "--V", 3 ) == 0 ||
                         strncmp ( argv[1], "--v", 3 ) == 0   ) )
     {
         printf ( "%s%.*s%s version %s (non-standalone)\n",
@@ -212,7 +248,22 @@ main ( int argc, char *argv[] )
     int                 fd_flags    = 0;
 
 
+#ifdef USE_PAM
+    {
+        /* Workaround a bug where if we wait until after all the daemonizing
+         * closing and opening of file handles, etc., pam_start may appear
+         * to succeed, but will fail to authenticate the user.
+         */
+        pam_handle_t    *pamdummy    = NULL;
+        struct pam_conv pamdummyconv = {NULL, NULL};
+        if ( pam_start ( USE_PAM, NULL, &pamdummyconv, &pamdummy) == PAM_SUCCESS)
+            pam_end ( pamdummy, 0 );
+    }
+#endif /* USE_PAM */
+
     if ( argc >= 2 && ( strncmp ( argv[1], "-v",  2 ) == 0 ||
+                        strncmp ( argv[1], "-V",  2 ) == 0 ||
+                        strncmp ( argv[1], "--V", 3 ) == 0 ||
                         strncmp ( argv[1], "--v", 3 ) == 0   ) )
     {
         printf ( "%s%.*s%s version %s (standalone)\n",
@@ -256,6 +307,7 @@ main ( int argc, char *argv[] )
     ptr = argv [ 1 ];
     if ( argc >= 2 && ( *ptr == ':' || isdigit ( (int) *ptr ) ) )
     {
+        int            j = 0;
         unsigned long  a = addr;
         unsigned short n = port;
         char           b [ 25 ] = "";
@@ -286,10 +338,11 @@ main ( int argc, char *argv[] )
          * We might have a port number
          */
         if ( ptr != NULL )
-            n = atoi ( ptr );
+            j = atoi ( ptr );
 
-        if ( a == BAD_ADDR || n == 0 || n > USHRT_MAX )
+        if ( a == BAD_ADDR || j == 0 || j > USHRT_MAX )
             err_dump ( HERE, "invalid address and/or port: \"%s\"", argv[1] );
+        n = j;
 
         port = htons ( n );
         addr = a;
@@ -317,13 +370,13 @@ main ( int argc, char *argv[] )
 #ifdef SYSLOG42
     openlog ( pname, 0 );
 #else
-    openlog ( pname, POP_LOGOPTS, /*LOG_DAEMON*/ POP_FACILITY );
+    openlog ( pname, POP_LOGOPTS, POP_FACILITY );
 #endif
 
     /*
-     * See if debug or trace options specified
+     * See if debug, trace, or log facility options specified
      */
-    i = getopt ( Qargc, Qargv, "b:BcCdD:e:f:FkK:l:L:p:RsSt:T:uUvy:" );
+    i = getopt ( Qargc, Qargv, "b:BcCdD:e:f:FkK:l:L:p:RsSt:T:uUvVxy:" );
     while ( i != EOF )
     {
         switch ( i )
@@ -335,19 +388,45 @@ main ( int argc, char *argv[] )
             case 't':
                 debug = TRUE;
                 trace_name = strdup ( optarg );
-                trace_file = fopen ( optarg, "a" );
+                trace_file = fopen  ( trace_name, "a" );
                 if ( trace_file == NULL )
-                    err_dump ( HERE, "Unable to open trace file \"%s\"", optarg );
+                    err_dump ( HERE, 
+                               "Unable to open trace file \"%.255s\"", 
+                               trace_name );
                 TRACE ( trace_file, POP_DEBUG, HERE,
-                        "Opened trace file \"%s\" as %d",
+                        "Opened trace file \"%.255s\" as %d",
                         trace_name, fileno(trace_file) );
                 break;
 
+            case 'y':  /* log facility */
+                {
+#ifndef    SYSLOG42
+                    POP config;         // these are dummy, just to pass to
+                    POP *p = &config;   // set_option; we'll do it for real
+                                        // later, in pop_init.
+                    if ( set_option ( p, "log-facility", 
+                                      optarg ) == POP_SUCCESS )
+                    {
+                        closelog();
+                        openlog ( pname, POP_LOGOPTS, p->log_facility );
+                    } 
+                    else 
+                    {
+                        fprintf ( stderr, 
+                                  "Error setting '-y' to %.255s\n", 
+                                  optarg );
+                        return 1;
+                    }
+#endif /* !SYSLOG42 */
+                }
+                break;
+
+
             default:
                 break;
         }
 
-        i = getopt ( Qargc, Qargv, "b:BcCdD:e:f:FkK:l:L:p:RsSt:T:uUvy:" );
+        i = getopt ( Qargc, Qargv, "b:BcCdD:e:f:FkK:l:L:p:RsSt:T:uUvVxy:" );
     }
     optind = 1; /* reset for pop_init */
 
@@ -446,10 +525,13 @@ main ( int argc, char *argv[] )
      * open files/process.  Then in a loop, the daemon can close all
      * possible file descriptors.
      */
-    TRACE ( trace_file, POP_DEBUG, HERE, "closing file descs %d to 0", sysconf ( _SC_OPEN_MAX )  );
+    TRACE ( trace_file, POP_DEBUG, HERE, 
+            "closing file descs %d to 0", 
+            sysconf ( _SC_OPEN_MAX )  );
     for ( i = sysconf ( _SC_OPEN_MAX ); i >= 0; i-- )
     {
-        if ( debug == FALSE || trace_file == NULL || i != fileno(trace_file) )
+        if ( debug == FALSE || trace_file == NULL 
+                            || i != fileno(trace_file) )
             close ( i );
     }
 
@@ -464,8 +546,8 @@ main ( int argc, char *argv[] )
      * that makes sense for your particular daemon.
      */
     i    = open ( "/dev/null",    O_RDONLY );           /* stdin  */
-    rslt = open ( "/dev/console", O_RDONLY );           /* stdout */
-    rslt = open ( "/dev/console", O_RDONLY );           /* stderr */
+    rslt = open ( "/dev/console", O_WRONLY );           /* stdout */
+    rslt = open ( "/dev/console", O_WRONLY );           /* stderr */
     if ( rslt > 0 )
         msg_out = err_out = fdopen ( rslt, "w+" );
     if ( msg_out == NULL )
@@ -483,7 +565,9 @@ main ( int argc, char *argv[] )
     sockfd = socket ( AF_INET, SOCK_STREAM, 0 );
     if ( sockfd < 0 )
         err_dump ( HERE, "Can't open stream socket" );
-    TRACE ( trace_file, POP_DEBUG, HERE, "opened stream socket; sockfd = %d", sockfd );
+    TRACE ( trace_file, POP_DEBUG, HERE, 
+            "opened stream socket; sockfd = %d", 
+            sockfd );
  
     i = 1;
     rslt = setsockopt ( sockfd, SOL_SOCKET, SO_REUSEADDR, 
@@ -500,7 +584,9 @@ main ( int argc, char *argv[] )
                     "%s: Server: setsockopt(SO_DEBUG) failed", pname );
     }
 
-    TRACE ( trace_file, POP_DEBUG, HERE, "set stream socket options; sockfd = %d", sockfd );
+    TRACE ( trace_file, POP_DEBUG, HERE, 
+            "set stream socket options; sockfd = %d", 
+            sockfd );
 
     bzero ( (char *) &serv_addr, sizeof(serv_addr) );
     serv_addr.sin_family      = AF_INET;
@@ -535,7 +621,9 @@ main ( int argc, char *argv[] )
           inet_ntoa ( serv_addr.sin_addr ),
           ntohs     ( serv_addr.sin_port ) );
 
-    TRACE ( trace_file, POP_DEBUG, HERE, "listening using socket fd %d", sockfd );
+    TRACE ( trace_file, POP_DEBUG, HERE, 
+            "listening using socket fd %d", 
+            sockfd );
 
     listen ( sockfd, 5 );
 
@@ -697,7 +785,7 @@ err_dump ( WHENCE, const char *format, ... )
     left  -= ( iChunk > 0 ? iChunk : left );
     len   += ( iChunk > 0 ? iChunk : left );
 
-    iChunk = Qsnprintf ( msg_buf + len, left, ": %s [%s:%d]",
+    iChunk = Qsnprintf ( msg_buf + len, left, ": %s [%s:%lu]",
                          sys_err_str(), fn, ln );
     left  -= ( iChunk > 0 ? iChunk : left );
     len   += ( iChunk > 0 ? iChunk : left );
@@ -759,7 +847,7 @@ err_msg ( WHENCE, const char *format, ... )
 
     va_end   ( ap );
 
-    iChunk = Qsnprintf ( msg_buf + len, left, ": %s [%s:%d]",
+    iChunk = Qsnprintf ( msg_buf + len, left, ": %s [%s:%lu]",
                          sys_err_str(), fn, ln );
     left  -= ( iChunk > 0 ? iChunk : left );
     len   += ( iChunk > 0 ? iChunk : left );
