@@ -1,10 +1,14 @@
 /*
- * Copyright (c) 2005 QUALCOMM Incorporated.  All rights reserved.
+ * Copyright (c) 2008 QUALCOMM Incorporated.  All rights reserved.
  * The file License.txt specifies the terms for use, modification,
  * and redistribution.
  *
  * Revisions:
  *
+ *     04/30/08  [rcg]
+ *             - Allow some options to have negative integer values,
+ *               specifically to allow announce-expire to be -1.
+ *
  *     04/11/05  [rcg]
  *             - Process user and spool config file as user, not as
  *               root (fix security hole reported by Jens Steube)
@@ -174,6 +178,7 @@ typedef enum
     kER_BAD_VAL,
     kER_EXP_BOOL,
     kER_EXP_INT,
+    kER_EXP_INT_NEG,
     kER_EXP_STR,
     kER_EXP_MNEM,
     kER_EXP_EOL,
@@ -277,7 +282,7 @@ typedef enum
 static config_table options [] =
 {
 /*    name                             type      restricted  value                */
-    { "announce-expire"              , CfgInt  , CfgResNone, kANNEXPIRE          },
+    { "announce-expire"              , CfgIntNeg,CfgResNone, kANNEXPIRE          },
     { "announce-login-delay"         , CfgInt  , CfgResNone, kANNLOGINDELAY      },
     { "auth-file"                    , CfgStr  , CfgResInit, kAUTHFILE           },
     { "auto-delete"                  , CfgBool , CfgResUser, kAUTO_DELETE        },
@@ -648,11 +653,18 @@ tilt ( POP              *p,
 
         case kER_EXP_INT:
             pop_log ( p, POP_PRIORITY, fn, ln, /* fn and ln are part of WHENCE */
-                      "Integer value expected for the %s option; "
+                      "Integer value (positive) expected for the %s option; "
                       "scanning \"%.*s\" at line %lu of config file %s",
                       ctp->name, MIN ( 128, len), token, (u_long)line_no, cfile );
             break;
 
+        case kER_EXP_INT_NEG:
+            pop_log ( p, POP_PRIORITY, fn, ln, /* fn and ln are part of WHENCE */
+                      "Integer value (may be negative) expected for the %s "
+                      "option; scanning \"%.*s\" at line %lu of config file %s",
+                      ctp->name, MIN ( 128, len), token, (u_long)line_no, cfile );
+            break;
+
         case kER_EXP_STR:
             pop_log ( p, POP_PRIORITY, fn, ln, /* fn and ln are part of WHENCE */
                       "String value expected for the %s option; "
@@ -1010,7 +1022,10 @@ get_token_length ( char *token, long line_length )
     char *orig_ptr = token;
 
 
-    if ( line_length > 0 && CharIsAlnum ( *token ) == FALSE )
+    if ( line_length == 0 )
+        return 0;
+
+    if ( *token != '-' && CharIsAlnum ( *token ) == FALSE )
         return 1; /* non-alphanumerics are their own token */
 
     while ( line_length > 0 && ( CharIsAlnum ( *token ) || *token == '-' ) )
@@ -1071,6 +1086,7 @@ show_result ( POP *p, config_table *opt, int iVal, const char *sVal, WHENCE )
             break;
 
         case CfgInt:
+        case CfgIntNeg:
             if ( p->debug )
                 pop_log ( p, POP_DEBUG, fn, ln, "Set %s to %d (%#0x)",
                           opt->name, iVal, iVal );
@@ -1126,6 +1142,21 @@ handle_value ( POP              *p,
     /*
      * Resolve the value
      */
+   if ( plen >= 2 && *pval == '-' && isdigit ( (int) *(pval +1) ) )
+    {
+        char *EndPtr = NULL;
+        iVal     = strtol ( pval, &EndPtr, 0 );
+        if ( iVal == LONG_MAX || iVal == LONG_MIN 
+                              || ( iVal == 0 && errno == EINVAL ) )
+        {
+            pop_log ( p, POP_PRIORITY, HERE,
+                      "Invalid value for %s: %.*s",
+                      opt_ent->name, plen, pval );
+            return kER_BAD_VAL;
+        }
+        val_type = CfgIntNeg;
+    }
+    else
     if ( isdigit ( (int) *pval ) )
     {
         char *EndPtr = NULL;
@@ -1280,6 +1311,18 @@ handle_value ( POP              *p,
             show_result ( p, opt_ent, *pOpt.pInt, NULL, HERE );
             break;
 
+        case CfgIntNeg:
+            if ( val_type != CfgInt && val_type != CfgIntNeg )
+                return kER_EXP_INT_NEG;
+
+            pOpt = get_opt_ptr ( p, opt_ent->value );
+            if ( pOpt.pInt == NULL )
+                return kER_INTERNAL;
+
+            *pOpt.pInt  = iVal;
+            show_result ( p, opt_ent, *pOpt.pInt, NULL, HERE );
+            break;
+
         case CfgMnem:
             if ( val_type != CfgStr )
                 return kER_EXP_MNEM;
@@ -1647,21 +1690,21 @@ check_config_files ( POP *p, struct passwd *pwp )
      */
     if ( bUser || bSpool ) {
         uid_save = geteuid();
-        if ( seteuid ( pwp->pw_uid ) != 0 ) {
-            rslt_usr = rslt_spl = POP_FAILURE; /* seteuid failed */
+        if ( SETEUID ( pwp->pw_uid ) != 0 ) {
+            rslt_usr = rslt_spl = POP_FAILURE; /* SETEUID failed */
             pop_log ( p, POP_PRIORITY, HERE,
-                      "seteuid(%i) for user %s failed",
+                      "SETEUID(%i) for user %s failed",
                       pwp->pw_uid, p->user );
-        } /* seteuid failed */
+        } /* SETEUID failed */
         else { /* we are now the user */
             if ( bUser )
                 rslt_usr = pop_config ( p, buf_usr, CfgUser );
             if ( bSpool )
                 rslt_spl = pop_config ( p, buf_spl, CfgUser );
 
-            if ( seteuid ( uid_save ) != 0 )
+            if ( SETEUID ( uid_save ) != 0 )
                 pop_log ( p, POP_PRIORITY, HERE,
-                          "seteuid(%i) back failed", uid_save );
+                          "SETEUID(%i) back failed", uid_save );
         } /* we are now the user */
     } /* bUser || bSpool */
 
