@@ -5,13 +5,28 @@
 
 
 /*
- * Copyright (c) 2000 Qualcomm Incorporated.  All rights reserved.
+ * Copyright (c) 2008 QUALCOMM Incorporated.  All rights reserved.
  * The file License.txt specifies the terms for use, modification,
  * and redistribution.
  *
  *
  * Revisions:
  *
+ *    4/11/05  [Randall Gellens]
+ *           - Open trace file as user, not root (fix security hole
+ *             reported by Jens Steube)
+ *           - Set umask to make sure we don't create group- or world-
+ *             writable files (e.g., trace or db)
+ *
+ *    7/29/03  [Randall Gellens]
+ *           - Removed prototype for srandom(); use stdlib.h instead.
+ *
+ *   01/02/00  [Randall Gellens]
+ *           - Added checks for dbm errors.
+ *
+ *   12/29/00  [Randall Gellens]
+ *           - Added -safe flag for -init.
+ *
  *    8/20/00  [Randall Gellens]
  *           - Added logging and tracing.
  *           - Reorganized command-line syntax.
@@ -80,13 +95,17 @@
 #  include <scram.h>
 #endif /* SCRAM */
 
-#ifdef GDBM
+#ifdef HAVE_GDBM_H
 #  include <gdbm.h>
-#else
-#  if HAVE_NDBM_H
+#else /* no GDBM */
+#  ifdef HAVE_NDBM_H
 #    include <ndbm.h>
-#  endif
-#endif /* GDBM */
+#  else /* no NDBM */
+#    ifdef HAVE_DBM_H
+#      include <dbm.h>
+#    endif /* HAVE_DBM_H */
+#  endif /* HAVE_NDBM_H */
+#endif /* HAVE_GDBM_H */
 
 #include <sys/types.h>
 #include <pwd.h>
@@ -116,8 +135,6 @@
 #  include <stdlib.h>
 #  define srandom srand
 #  define  random  rand
-#else
-   extern void srandom();
 #endif /* HAVE_SRANDOM */
 
 #include <flock.h>
@@ -144,6 +161,14 @@ extern int decode64();
 char *strerror();
 #endif /* HAVE_STRERROR */
 
+#ifdef HAVE_SYS_UNISTD_H
+#  include <sys/unistd.h>
+#endif /* HAVE_SYS_UNISTD_H */
+
+#ifdef HAVE_UNISTD_H
+#  include <unistd.h>
+#endif /* include <unistd.h> */
+
 #include "popper.h"
 #include "logit.h"
 
@@ -166,6 +191,7 @@ static struct mods {
 
 typedef enum {
     INITSW  = 1,
+    SAFESW,
     LISTSW,
     USERSW,
     DELESW,
@@ -174,6 +200,7 @@ typedef enum {
     TRACESW,
     DEBUGSW,
     VERSNSW,
+    DUMPSW,
 
     BAD_SW
 } switchType;
@@ -185,6 +212,7 @@ static struct swit {
     switchType val;
 } switches[] = {
     { "init"    , 0, INITSW  },
+    { "safe"    , 0, SAFESW  },
     { "list"    , 0, LISTSW  },
     { "user"    , 1, USERSW  },
     { "delete"  , 1, DELESW  },
@@ -194,8 +222,11 @@ static struct swit {
     { "trace"   , 1, TRACESW },
     { "debug"   , 0, DEBUGSW },
     { "version" , 0, VERSNSW },
+#ifdef _DEBUG
+    { "dump"    , 0, DUMPSW  },
+#endif /* _DEBUG */
 
-    { NULL        },
+    { NULL      , 0, BAD_SW  },
 };
 
 
@@ -253,6 +284,17 @@ static int     debug        =    0;
 static FILE  *trace_file    = NULL;
 
 
+/*
+ * Prototypes
+ */
+static void         byebye  ( int val );
+static void         adios   ( WHENCE, const char *fmt, ... );
+static void         helpful ( void );
+static int          check_db_err ( void *db, const char *op, BOOL bExp );
+static const char  *printable ( const char *p, int len );
+static void         open_trace ( char *fname );
+
+
 static void
 byebye ( int val )
 {
@@ -288,15 +330,168 @@ static void
 helpful ( void )
 {
     fprintf  ( stderr, "usage: %s [ -debug | -trace <tracefile> ] [ <action> ]\n"
-                       "<action>: -init | -list [<user> | ALL] | -only | "
-                       "-user <user> [ <password> ] | \n"
-                       "          -delete <user> | -help | -version\n",
+                       "<action>: -init [ -safe ] \n"
+                       "          -list [ <user> | ALL ] | -only\n"
+                       "          -user <user> [ <password> ]\n"
+                       "          -delete <user>\n"
+                       "          -help\n"
+                       "          -version\n",
                program );
 
     byebye ( 1 );
 }
 
 
+/*
+ * Check for db errors.
+ *
+ * Parameters:
+ *     db:          result of *dbm_open.
+ *     op:          text string of operation (e.g., "fetch").
+ *     bExp:        if an error (such as "item not found" is
+ *                      expected and not anything to get excited
+ *                      about.
+ */
+static int
+check_db_err ( void *db, const char *op, BOOL bExp )
+{
+    int   db_err = 0;
+    char *db_str = NULL;
+    int   sv_err = errno;
+    int   vNotFound = -1;
+    int   vEmptyDb  = -1;
+
+
+#ifdef GDBM
+    (void) db;
+    db_err    = (int) gdbm_errno;
+    db_str    = (char *) gdbm_strerror ( gdbm_errno );
+    vNotFound = GDBM_ITEM_NOT_FOUND;
+    vEmptyDb  = GDBM_EMPTY_DATABASE;
+#else
+    db_err = dbm_error  ( (DBM *) db );
+    db_str = strerror   ( db_err );
+#endif /* GDBM */
+
+    if ( db_err != 0 ) {
+        if ( bExp && ( db_err == vNotFound || db_err == vEmptyDb ) ) {
+            BLATHER5 ( "DB error on %s: %s %d (%d: %s)",
+                       op, db_str, db_err, sv_err, strerror(sv_err) );
+        } else {
+            logit ( trace_file, POP_NOTICE, HERE,
+                    "*** DB error on %s: %s %d (%d: %s) ***",
+                    op, db_str, db_err, sv_err, strerror(sv_err) );
+
+            fprintf ( stderr, "*** DB error on %s: %s %d (%d: %s) ***\n",
+                      op, db_str, db_err, sv_err, strerror(sv_err) );
+        }
+    } else {
+        BLATHER5 ( "DB %s succeeded: %s %d (%d: %s)",
+                   op, db_str, db_err, sv_err, strerror(sv_err) );
+    }
+
+    return db_err;
+}
+
+
+static char printable_buffer [ 1024 ];
+
+static const char *
+printable ( const char *p, int len )
+{
+    int     left = sizeof ( printable_buffer ) -1;
+    char   *q    = printable_buffer;
+    char    c    = 0;
+
+
+    if ( p == NULL ) {
+        BLATHER0 ( "printable passed NULL pointer" );
+        return NULL;
+    }
+
+    /*BLATHER2 ( "printable passed %d (strlen %d)", len, strlen(p) );*/
+
+    while ( left > 0 && len > 0 ) {
+        c = *p++;
+        len--;
+        if ( c >= ' ' && c <= '~' ) {
+            *q++  = c;
+            left -= 1;
+        } else {
+            long ic = c & 0x000000ff;
+            switch ( c ) {
+                case '\0':
+                    *q++  = '\\';
+                    *q++  = '0';
+                    left -= 2;
+                    break;
+                case '\n':
+                    *q++  = '\\';
+                    *q++  = 'n';
+                    left -= 2;
+                    break;
+                case '\t':
+                    *q++  = '\\';
+                    *q++  = 't';
+                    left -= 2;
+                    break;
+                case '\r':
+                    *q++  = '\\';
+                    *q++  = 'r';
+                    left -= 2;
+                    break;
+                case '\a':
+                    *q++  = '\\';
+                    *q++  = 'a';
+                    left -= 2;
+                    break;
+                case '\b':
+                    *q++  = '\\';
+                    *q++  = 'b';
+                    left -= 2;
+                    break;
+                case '\v':
+                    *q++  = '\\';
+                    *q++  = 'v';
+                    left -= 2;
+                    break;
+                default:
+                    sprintf ( q, "\\%#lx", ic );
+                    q    += 5;
+                    left -= 5;
+            } /* switch */
+        } /* unprintable */
+    } /* while */
+
+    *q = '\0';
+    /*BLATHER2 ( "printable returning %d: '%s'", strlen(printable_buffer), printable_buffer );*/
+    return printable_buffer;
+}
+
+void
+open_trace ( char *tname )
+{
+    UID_T uid_save = -1;
+    UID_T myuid    = -1;
+
+
+    uid_save = geteuid();
+    myuid = getuid();
+    if ( SETEUID ( myuid ) != 0 )
+        adios ( HERE, "internal error @ %i", __LINE__ );
+    
+    trace_file = fopen ( tname, "a+" );
+    if ( trace_file == NULL )
+        adios ( HERE, "Unable to open trace file \"%s\": %s (%d)\n",
+                tname, STRERROR(errno), errno );
+    BLATHER1 ( "Trace and Debug destination is file \"%s\"",
+               tname );
+            
+    if ( SETEUID ( uid_save ) != 0 )
+        adios ( HERE, "internal error @ %i", __LINE__ );
+}
+
+
 #ifndef HAVE_STRDUP
 #include <stddef.h>
 
@@ -359,14 +554,6 @@ char *getpass();
 #  include <stdio.h>
 #  include <termios.h>
 
-#  ifdef HAVE_SYS_UNISTD_H
-#    include <sys/unistd.h>
-#  endif /* HAVE_SYS_UNISTD_H */
-
-#  ifdef HAVE_UNISTD_H
-#    include <unistd.h>
-#  endif /* include <unistd.h> */
-
 #  ifndef TCSASOFT
 #    define TCSASOFT 0
 #  endif
@@ -460,23 +647,27 @@ main ( argc, argv )
 int argc;
 char   *argv[];
 {
-    UID_T          myuid;
+    UID_T          myuid                    = -1;
     int            flags                    = 0,
                    i                        = 0,
                    apoplen                  = 0,
                    scramlen                 = 0;
     BOOL           delesw                   = FALSE,
                    initsw                   = FALSE,
+                   safesw                   = FALSE,
                    onlysw                   = FALSE,
                    passtype                 = FALSE,
                    insist                   = FALSE,
                    listsw                   = FALSE,
                    usersw                   = FALSE,
-                   popuser                  = FALSE;
+                   dumpsw                   = FALSE,
+                   popuser                  = FALSE,    /* TRUE: we are the db owner */
+                   user_found               = FALSE;    /* TRUE: user to be changed exists */
     switchType     v                        = BAD_SW;
     char          *cp                       = NULL,
                   *userid                   = NULL,
                   *givenpassword            = NULL,
+                  *db_name                  = NULL,
                    apopfld       [ BUFSIZ ] = "",
                    scramfld      [ BUFSIZ ] = "",
                    outfld        [ BUFSIZ ] = "",
@@ -509,6 +700,7 @@ char   *argv[];
 
     memset  ( &pop_pw, 0, sizeof(pop_pw) );
     memset  ( &my_pw,  0, sizeof(my_pw)  );
+    umask   ( 0077 ); /* make sure we don't create group- or world-writable files */
     srandom ( (unsigned int) time ( (TIME_T *) 0) );   /* seed random with the
                                                           current time */
 
@@ -588,13 +780,7 @@ char   *argv[];
                 helpful();
             case TRACESW:
                 debug++;
-                trace_file = fopen ( argv[1], "a+" );
-                if ( trace_file == NULL )
-                    adios ( HERE, 
-                            "Unable to open trace file \"%s\": %s (%d)\n",
-                            argv[1], STRERROR(errno), errno );
-                BLATHER1 ( "Trace and Debug destination is file \"%s\"",
-                           argv[1] );
+                open_trace ( argv[1] );
                 argc--;
                 argv++;
                 break;
@@ -610,6 +796,11 @@ char   *argv[];
                 listsw = FALSE;
                 delesw = FALSE;
                 break;
+            case SAFESW:
+                safesw = TRUE;
+                break;
+            case DUMPSW:
+                dumpsw = TRUE;
             case LISTSW:
                 listsw = TRUE;
                 initsw = FALSE;
@@ -655,8 +846,8 @@ char   *argv[];
 /* Expand the rest iff SCRAM or APOP and favor SCRAM to hold auth db name */
 
 #ifdef   SCRAM
-#define  AUTHON
-#define  AUTHDB SCRAM
+#  define  AUTHON
+#  define  AUTHDB SCRAM
 #else
 #  ifdef   APOP
 #    define  AUTHON
@@ -699,8 +890,8 @@ char   *argv[];
                 "Only superuser or user \"%s\" can perform the requested function",
                 POPUID );
 
-    if ( myuid != 0 && initsw )
-        adios ( HERE, "Only superuser can initialize the authentication database" );
+    if ( myuid != 0 && ( initsw || dumpsw ) )
+        adios ( HERE, "Only superuser can perform the requested function" );
 
     if ( delesw )
         fprintf ( stderr, "Warning: deleting user \"%s\"\n",
@@ -709,16 +900,24 @@ char   *argv[];
 #ifdef GDBM
     strlcpy  ( auth_file, AUTHDB, sizeof(auth_file) - 1 );
     BLATHER1 ( "GDBM: auth_file=%s", auth_file );
-#else
+    db_name = auth_file;
+#else /* not GDBM */
     strlcpy  ( auth_dir, AUTHDB, sizeof(auth_dir) - 5 );
 #  ifdef BSD44_DBM
     strcat   ( auth_dir, ".db" );
     BLATHER1 ( "BSD44_DBM: auth_dir=%s", auth_dir );
-#  else
+    db_name = auth_dir;
+#  else /* non-BSD44 DBM */
     strlcpy  ( auth_pag, AUTHDB, sizeof(auth_pag) - 5 );
     strcat   ( auth_pag, ".pag" );
     strcat   ( auth_dir, ".dir" );
     BLATHER2 ( "xDBM: auth_pag='%s'; auth_dir='%s'", auth_pag, auth_dir );
+    db_name = malloc ( sizeof ( auth_pag ) + 15 );
+    if ( db_name != NULL ) {
+        strcpy ( db_name, auth_pag );
+        strcat ( db_name, " (and .dir )" );
+    } else
+        db_name = auth_pag;
 #  endif /* BSD44_DBM */
 #endif /* GDBM */
 
@@ -739,12 +938,17 @@ char   *argv[];
 
         key.dptr  = userid;
         key.dsize = strlen ( key.dptr ) + 1;
+        BLATHER2 ( "...fetching: (%d) '%s'",
+                   key.dsize, printable ( key.dptr, key.dsize ) );
 
 #ifdef GDBM
         value = gdbm_fetch ( db, key );
 #else
         value = dbm_fetch  ( db, key );
 #endif
+
+        check_db_err ( db, "fetch", TRUE );
+
         if ( value.dptr == NULL )
             adios ( HERE, "User '%s' not found in authentication database", userid );
 
@@ -780,7 +984,13 @@ char   *argv[];
 #endif
         {
             char ibuf [ 30 ];
-        
+
+            if ( safesw == TRUE ) {
+                fprintf  ( stderr, "POP authentication DB not "
+                                   "initialized (exists): %s\n",
+                           db_name );
+                byebye ( 0 );
+            }
             printf ( "Really initialize POP authentication DB? " );
             if ( fgets ( ibuf, sizeof(ibuf), stdin ) == NULL || ibuf[0] != 'y' )
                 byebye ( 1 );
@@ -863,30 +1073,47 @@ char   *argv[];
         if ( strcmp ( userid, "ALL" ) != 0 ) { /* list one user */
             key.dptr  = userid;
             key.dsize = strlen ( key.dptr ) + 1;
+            BLATHER2 ( "...fetching: (%d) '%s'",
+                       key.dsize, printable ( key.dptr, key.dsize ) );
+
 #ifdef GDBM
             value = gdbm_fetch ( db, key );
 #else
             value = dbm_fetch  ( db, key );
 #endif
+
+            check_db_err ( db, "fetch", TRUE );
+
             if ( value.dptr == NULL )
                 adios ( HERE, "no entry for \"%s\" in POP authentication DB", userid );
 
-            (strncpy(buf, key.dptr, sizeof(buf)-2))[sizeof(buf)-2] = '\0';
-            strcat ( buf, ":" );
-            printf ( "%-16s", buf );
-            if ( ( (char *) value.dptr )[0] != '\0' ) 
-                printf ( " APOP" );
-            i = strlen(value.dptr) + 2;
-            if ( value.dsize >= i
-                    && ( (char *) value.dptr )[i] != '\0' )
-                printf ( " SCRAM" );
-            printf ( "\n" );
+            strcpy ( apopfld, value.dptr ); 
+            if ( (size_t) value.dsize > strlen(apopfld) + 1 )
+                strcpy ( scramfld, &((char *) value.dptr) [strlen(apopfld) + 1] );
+            else
+                scramfld[0] = '\0';
+            apoplen  = strlen (  apopfld ) + 1;
+            scramlen = strlen ( scramfld ) + 1;
+
+            if ( dumpsw ) {
+                printf ( "%.16s:%*s %.62s\n",
+                         key.dptr,
+                         16 - (int)strlen(key.dptr), " ",
+                         printable ( value.dptr, value.dsize ) );
+                BLATHER5 ( "Dumping \"%s\": (%d)'%s'; APOP %d; SCRAM %d",
+                           key.dptr,
+                           value.dsize, printable ( value.dptr, value.dsize ),
+                           apoplen, scramlen );
+            } else {
+                printf ( "%-16s: %s %s\n", key.dptr,
+                         ( apoplen  > 1 ) ? "APOP"  : "    ",
+                         ( scramlen > 1 ) ? "SCRAM" : ""    );
+                BLATHER3 ( "Listing user \"%s\": APOP %s; SCRAM %s",
+                           key.dptr,
+                           ( apoplen  > 1 ) ? "yes" : "no",
+                           ( scramlen > 1 ) ? "yes" : "no" );
+            }
 
-            BLATHER3 ( "Listing user \"%s\": APOP %s; SCRAM %s",
-                       buf,
-                       ( ( (char *) value.dptr )[0] != '\0' ) ? "yes" : "no",
-                       ( value.dsize >= i &&
-                         ( (char *) value.dptr)[i] != '\0' )  ? "yes" : "no" );
         } /* list one user */
     else { /* list all users */
         BLATHER0 ( "Listing all users... " );
@@ -896,34 +1123,58 @@ char   *argv[];
         for ( key = dbm_firstkey  ( db ); key.dptr; key = dbm_nextkey  ( db      ) ) 
 #endif
         {
-            (strncpy(buf, key.dptr, sizeof(buf)-2))[sizeof(buf)-2] = '\0';
-            strcat ( buf, ":" );
-            printf ( "%-16s", buf );
+            BLATHER2 ( "...fetching: (%d)'%s'",
+                       key.dsize, printable ( key.dptr, key.dsize ) );
+
 #ifdef GDBM
             value = gdbm_fetch ( db, key );
 #else
             value = dbm_fetch  ( db, key );
 #endif
+
+            check_db_err ( db, "fetch", FALSE );
+
             if ( value.dptr == NULL ) {
-                printf   ( " *** no information?!?\n" );
+                printf   ( "%.16s:%*s *** no information?!?\n",
+                           key.dptr, 16 - (int)strlen(key.dptr), " " );
                 BLATHER1 ( "*** No information for user \"%s\"", buf );
             }
             else {
-                if ( ( (char *) value.dptr )[0] != '\0' )
-                    printf ( " APOP" );
-                i = strlen ( value.dptr ) + 2;
-                if ( value.dsize >= i
-                           && ( (char *) value.dptr )[i] != '\0' )
-                    printf ( " SCRAM" );
-                printf ( "\n" );
+                strcpy ( apopfld, value.dptr ); 
+                if ( (size_t) value.dsize > strlen(apopfld) + 1 )
+                    strcpy ( scramfld, &((char *) value.dptr) [strlen(apopfld) + 1] );
+                else
+                    scramfld[0] = '\0';
+                apoplen  = strlen (  apopfld ) + 1;
+                scramlen = strlen ( scramfld ) + 1;
+
+                if ( dumpsw ) {
+                    printf ( "%.16s:%*s %.62s\n",
+                             key.dptr,
+                             16 - (int)strlen(key.dptr), " ",
+                             printable ( value.dptr, value.dsize ) );
+                    BLATHER5 ( "Dumping \"%s\": (%d)'%s'; APOP %d; SCRAM %d",
+                               key.dptr,
+                               value.dsize, printable ( value.dptr, value.dsize ),
+                               apoplen, scramlen );
+                } else {
+                    printf ( "%-16s: %s %s\n", key.dptr,
+                             ( apoplen  > 1 ) ? "APOP"  : "    ",
+                             ( scramlen > 1 ) ? "SCRAM" : ""    );
+                    BLATHER3 ( "Listing user \"%s\": APOP %s; SCRAM %s",
+                               key.dptr,
+                               ( apoplen  > 1 ) ? "yes" : "no",
+                               ( scramlen > 1 ) ? "yes" : "no" );
+                }
 
-                BLATHER3 ( "...user \"%s\": APOP %s; SCRAM %s",
-                           buf,
-                           ( ( (char *) value.dptr )[0] != '\0' ) ? "yes" : "no",
-                           ( value.dsize >= i &&
-                             ( (char *) value.dptr)[i] != '\0' )  ? "yes" : "no" );
             }
         } /* for loop */
+
+        /*
+         * Check if we terminated normally or on an error
+         */
+        check_db_err ( db, "first/next key", TRUE );
+        
     } /* list all users */
 
 #ifdef GDBM
@@ -935,43 +1186,63 @@ char   *argv[];
     byebye ( 0 );
     } /* listsw */
 
+    /*
+     * If we're here, we must be doing a password change (usersw)
+     */
     if ( userid == NULL ) {
         userid = my_pw.pw_name;
     } else {
         pw = getpwnam ( userid );
         if ( pw == NULL )
-            adios ( HERE, "Sorry, don't know who uid %s is.", userid );
+            adios ( HERE, "Sorry, don't know who user %s is.", userid );
         userid = pw->pw_name;
     }
 
-    fprintf ( stderr, "Changing %s%s %s for %s.\n",
-              onlysw ? "only " : "",  
-              (mode == SCRAM_AUTH) ? "SCRAM" : "APOP",
-              passtype ? "pass phrase" : "password",
-              userid );
-
     key.dptr  = userid;
     key.dsize = strlen ( key.dptr ) + 1;
+    BLATHER2 ( "...fetching: (%d) '%s'",
+               key.dsize, printable ( key.dptr, key.dsize ) );
+
 #ifdef GDBM
     value = gdbm_fetch ( db, key );
 #else
     value = dbm_fetch  ( db, key );
 #endif
-    
-    if ( value.dptr != NULL ) {
+
+    check_db_err ( db, "fetch", TRUE );
+    user_found = ( value.dptr != NULL );
+
+    if ( user_found ) {
         strcpy ( apopfld, value.dptr ); 
         if ( (size_t) value.dsize > strlen(apopfld) + 1 )
             strcpy ( scramfld, &((char *) value.dptr) [strlen(apopfld) + 1] );
         else
             scramfld[0] = '\0';
+
+        BLATHER3 ( "user %s exists in db (%d/%d)",
+                   userid, strlen(apopfld), strlen(scramfld) );
     }
-    else 
+    else {
         apopfld[0] = scramfld[0] = '\0';
+        BLATHER1 ( "user %s doesn't exist -- adding", userid );
+    }
+
+    fprintf ( stderr, "%s %s%s %s for %s.\n",
+              user_found           ? "Changing"     : "Adding",
+              onlysw               ? "only "        : "",  
+              (mode == SCRAM_AUTH) ? "SCRAM"        : "APOP",
+              passtype             ? "pass phrase"  : "password",
+              userid );
 
     apoplen  = strlen (  apopfld ) + 1;
     scramlen = strlen ( scramfld ) + 1;
 
-    if ( (myuid && !popuser) 
+    /*
+     * If user isn't root or the db owner, and user has an existing
+     * password for the mode (apop or scarm) being changed, verify old
+     * password.
+     */
+    if ( ( myuid && !popuser ) 
           && (((mode==APOP_AUTH || mode==POP_AUTH) && apopfld[0] )
                 ||(mode==SCRAM_AUTH && scramfld[0] )) ) {
 
@@ -1007,7 +1278,8 @@ char   *argv[];
 #endif
                 adios ( HERE, "Sorry, pass phrase entered incorrectly\n");
         }
-    }
+    } /* old password verification */
+
 #ifdef GDBM
     gdbm_close ( db );
 #else
@@ -1027,6 +1299,10 @@ char   *argv[];
                                 : ( passtype ? "a longer pass phrase"
                                              : "a longer password" ) );
 
+        /*
+         * If new password passed in (givenpassword) use it,
+         * otherwise prompt for the new one.
+         */
         if ( givenpassword != NULL ) {
             strncpy ( buf, givenpassword, sizeof ( buf ) );
             buf [ sizeof(buf) -1 ] = '\0';
@@ -1170,24 +1446,32 @@ char   *argv[];
     value.dsize  = strlen ( value.dptr = strcpy(outfld, apopfld) ) + 1;
     value.dsize += strlen ( strcpy(&outfld[value.dsize], scramfld) ) + 1;
 
-#ifdef GDBM
-    if ( gdbm_store ( db, key, value, GDBM_REPLACE ) )
-        adios ( HERE, "POP authentication DB may be corrupt?!?" );
+#  ifdef GDBM
+    i = gdbm_store ( db, key, value, GDBM_REPLACE );
+    check_db_err ( db, "store", FALSE );
     gdbm_close ( db );
-#else /* not GDBM */
-    if ( dbm_store ( db, key, value, DBM_REPLACE ) )
-        adios ( HERE, "POP authentication DB may be corrupt?!?" );
+#  else /* not GDBM */
+    i = dbm_store  ( db, key, value, DBM_REPLACE );
+    check_db_err ( db, "store", FALSE );
     dbm_close ( db );
-#endif /* GDBM */
+#  endif /* GDBM */
+
+    check_db_err ( db, "close", FALSE );
+
+    if ( i )
+        adios ( HERE, "POP authentication DB %s may be corrupt?!?", db_name );
+
 #endif /* AUTHON */
 
     logit  ( trace_file, POP_NOTICE, HERE,
-             "popauth: %s changed %s%s %s for %s",
+             "popauth: %s %s %s%s %s for %s",
              my_pw.pw_name,
-             onlysw ? "only " : "",  
-             (mode == SCRAM_AUTH) ? "SCRAM" : "APOP",
-             passtype ? "pass phrase" : "password",
+             user_found             ? "changed"     : "added",
+             onlysw                 ? "only "       : "",  
+             (mode == SCRAM_AUTH)   ? "SCRAM"       : "APOP",
+             passtype               ? "pass phrase" : "password",
              userid );
+
     byebye ( 0 );
     return 0; /* avoid a warning */
 }
