@@ -5,7 +5,7 @@
  */
  
 /*
- * Copyright (c) 2005 QUALCOMM Incorporated.  All rights reserved.
+ * Copyright (c) 2008 QUALCOMM Incorporated.  All rights reserved.
  * The file License.txt specifies the terms for use, modification,
  * and redistribution.
  */
@@ -192,6 +192,15 @@ typedef enum {
 #  define EPROTO ECONNABORTED
 #endif /* no EPROTO */
 
+/*
+ * Some OSes (notably HPUX) use setresuid() instead of seteuid()
+ */
+#ifdef HAVE_SETRESUID
+#  define SETEUID(E) setresuid(-1,E,-1)
+#else
+#  define SETEUID(E) seteuid(E)
+#endif /* HAVE_SETRESUID */
+
 /*
  * In standalone mode we need to call _exit rather than exit
  */
@@ -415,7 +424,8 @@ typedef enum {
  */
 typedef enum {                               /* Config file option is:         */
     CfgBool = 1,                             /* - boolean                      */
-    CfgInt,                                  /* - integer                      */
+    CfgInt,                                  /* - integer (positive)           */
+    CfgIntNeg,                               /* - integer, may be negative     */
     CfgMnem,                                 /* - mnemonic                     */
     CfgStr,                                  /* - string                       */
     CfgBad                                   /* - INVALID VALUE                */
@@ -857,7 +867,7 @@ typedef struct {
                                  in pop_extend.c. Used by pop_parse
                                  for expansion. */
 enum { MAX_MDEF_INDEX = 10 };
-enum { MAX_MDEF_NAME  = 100 };
+enum { MAX_MDEF_LENGTH = 100 };
 
 
 MsgInfoList *msg_ptr ( POP *p, int msg_num );
