@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2000 QUALCOMM Incorporated.  All rights reserved.
+ * Copyright (c) 2008 QUALCOMM Incorporated.  All rights reserved.
  * The license.txt file specifies the terms and conditions for
  * modifications and redistribution.
  *
@@ -13,7 +13,8 @@
  *  Current version of Qpopper
  */
 
-#define VERS_NUM          "3.1"
+
+#define VERS_NUM          "4.0.12"
 
 #ifdef KERBEROS
 #  ifdef KRB4
